/*
 * ftc.c
 *
 * Fine Tooth Comb
 *
 * Copyright (c) 2003 Bob Fleck <bob@shmoo.com>
 * All rights reserved.
 *
 * Portions of HCI code derived from the FreeBSD hccontrol utility:
 * Copyright (c) 2001-2002 Maksim Yevmenkin <m_evmenkin@yahoo.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "ftc.h"
#include "ftc_manuf.h"

int scanning = 1;

/* Signal handlers that just stop the program loop and let it finish up */
void handle_sigint(int sig)  { scanning = 0; }
void handle_sigterm(int sig) { scanning = 0; }
void handle_sigquit(int sig) { scanning = 0; }

/* Use sigaction for signal handlers if it's available */
static void setup_signal_handler(int sig, void (*handler)()) {
#if _POSIX_VERSION > 198800L
  struct sigaction action;

  action.sa_handler = handler;
  sigemptyset(&(action.sa_mask));
  sigaddset(&(action.sa_mask), sig);
  action.sa_flags = 0;
  sigaction(sig, &action, 0);
#else
  signal(sig, handler);
#endif
} /* setup_signal_handler */

/* Update/append entries in inquiry response collection */
static int inqr_update_unique(bdaddr_t *res, int resn, ng_hci_inquiry_result_ep *ir) {
  int found;
  int i, j;
  u_int8_t *pi = (u_int8_t *)(ir + 1);
  struct inquiry_response *c_ir = (struct inquiry_response *)pi;

  /* Don't let this run up the stack
   * I didn't check (resn + ir->num_responses because of what this func
   * is doing.
   * resn <= returned resn <= resn+ir->num_responses
   * So, check to make sure it's not full already, and then check after
   * each append to make sure we haven't filled it.
   */
  if (resn >= FTC_MAX_INQR) return FTC_MAX_INQR;

  if (resn == 0) {
    /* First result, so just copy */
    memcpy(&res[0], &c_ir->bdaddr, sizeof(res[0]));
    resn++;
  } else {
    for (i = 0; i < ir->num_responses; i++) {
      found = 0;
      for (j = 0; j < resn; j++) {
	c_ir = (struct inquiry_response *)pi;     
	
	/* If the addrs match, we're done with this response */
	if (memcmp(&res[j], &c_ir->bdaddr, sizeof(res[j])) == 0) {
	  found = 1;
	  break;
	}
      }
      
      /* Append the address if it wasn't there already*/
      if(!found) memcpy(&res[resn++], &c_ir->bdaddr, sizeof(res[j]));      

      /* Have to check again here to make sure stack doesn't get smashed */
      if (resn >= FTC_MAX_INQR) return FTC_MAX_INQR;
    
      /* Go to next entry in inquiry result */
      pi += sizeof(*c_ir);
    }
  }
  
  return resn;
} /* inqr_update_unique */

/* Print out all the received responses */
static void inqr_print(bdaddr_t *res, int resn) {
  int i;

  for (i = 0; i < resn; i++) {
    printf("++>IR>%02x:%02x:%02x:%02x:%02x:%02x\n",
	   res[i].b[5], res[i].b[4], res[i].b[3],
	   res[i].b[2], res[i].b[1], res[i].b[0]);
  }
} /* inqr_print */

/* Print the packet for someone trying to connect to us */
static void cr_print(ng_hci_con_req_ep *cr) {
  char type;
  
  if (cr->link_type == 0x00) type = 'S'; /* SCO connection */
  else if (cr->link_type == 0x01) type = 'A'; /* ACL connection */
  else type = '?';

  printf("++>CR>%02x:%02x:%02x:%02x:%02x:%02x>%c>%02x%02x%02x\n",	 
	 cr->bdaddr.b[5], cr->bdaddr.b[4], cr->bdaddr.b[3],
	 cr->bdaddr.b[2], cr->bdaddr.b[1], cr->bdaddr.b[0],
	 type, cr->uclass[2], cr->uclass[1], cr->uclass[0]);
} /* cr_print */

/* Print out a brute force status message */
static void bf_print(bdaddr_t *bdaddr, char status){
  printf("%c%c>BF>%02x:%02x:%02x:%02x:%02x:%02x\n",
	 status, status,
	 bdaddr->b[5], bdaddr->b[4], bdaddr->b[3],
	 bdaddr->b[2], bdaddr->b[1], bdaddr->b[0]);
} /* bf_print */

/* Step to next manufacturer from ff_manuf.h */
static void next_manufacturer(bdaddr_t *bfaddr) {
  static int cm = 0;
  static int totalm = SIZE(ftc_manuf);

  bfaddr->b[5] = ftc_manuf[cm][0];
  bfaddr->b[4] = ftc_manuf[cm][1];
  bfaddr->b[3] = ftc_manuf[cm][2];
  bfaddr->b[2] = 0x00;
  bfaddr->b[1] = 0x00;
  bfaddr->b[0] = 0x00;
  if (++cm > totalm) cm = 0;
} /* next_manufacturer */

/* Increment the BDADDR, starting with last byte */
static void next_bdaddr(bdaddr_t *bfaddr) {
  /* Increment the MAC */
  if (bfaddr->b[0]++ == 0xFF) {
    bfaddr->b[0] = 0x00;
    if (bfaddr->b[1]++ == 0xFF) {
      bfaddr->b[1] = 0x00;
      if (bfaddr->b[2] == 0xFF) {
	next_manufacturer(bfaddr);
      }
    }
  }
} /* next_bdaddr */

/* Combined periodic inquiry scan with brute force */
static int hci_scan(int s, int brute)
{
  int n0, i, pause = 0;
  u_int8_t b[512];
  /* Array of BDADDRs that responded to inquiry */
  bdaddr_t inqres[FTC_MAX_INQR];
  int inqresn = 0;
  /* Current bdaddr being brute forced */
  bdaddr_t bfaddr;

  /* Response packets - just casting these for readability below */
  ng_hci_event_pkt_t *e = (ng_hci_event_pkt_t *) b;
  ng_hci_con_compl_ep *cc = (ng_hci_con_compl_ep *)(e+1);
  ng_hci_discon_compl_ep *dc = (ng_hci_discon_compl_ep *)(e+1);
  ng_hci_command_compl_ep *ccomp = (ng_hci_command_compl_ep *)(e+1);
  u_int8_t *ccomp_result = (u_int8_t *)(ccomp+sizeof(*ccomp));
  ng_hci_command_status_ep *cstat = (ng_hci_command_status_ep *)(e+1);
  ng_hci_con_req_ep *cr = (ng_hci_con_req_ep *)(e+1);

  /* Kick off periodic inquiry on GIAC LAP */
  hci_periodic_inq_start(s, 0x33);

  if (brute && scanning) {
    /* Set it to first manufacturer */
    next_manufacturer(&bfaddr);
    hci_connect(s, &bfaddr);
  }

  while(scanning) {
    /* Loop for events */
    n0 = sizeof(b);
    if (hci_recv_pkt(s, b, n0) == -1) {
      continue;
    }
    
    switch (e->event) {
      case NG_HCI_EVENT_INQUIRY_RESULT:
	/* Received multiple times per periodic inquiry */
	inqresn = inqr_update_unique(inqres, inqresn, (ng_hci_inquiry_result_ep *)(e + 1));
	break;
      
      case NG_HCI_EVENT_INQUIRY_COMPL:
	/* Receive after each periodic inquiry complete */
	inqr_print(inqres, inqresn);
	inqresn = 0;
	break;
	
      case NG_HCI_EVENT_CON_COMPL:
	/* Received upon completion of connection attempt */
	/* Ignore our rejects of other conn attempts */
	if (cc->status != FTC_ERR_REJECT) {
	  if (cc->status == 0x00) {
	    /* Last connect succeeded, drop it*/
	    if(!pause) hci_disconnect(s, cc->con_handle);
	  }
	  
	  bf_print(&bfaddr,((cc->status == 0x00) ? '+' : '-'));
	  
	  /* Now try the next */
	  if ((!pause) && brute) {
	    next_bdaddr(&bfaddr);
	    hci_connect(s, &bfaddr);
	  }
	}
	break;
	
      case NG_HCI_EVENT_CON_REQ:
	/* Someone is trying to connect to us and making life easy :)
	 * Thanks to gdead for suggesting this.
	 */
	cr_print(cr);
	/* Reject it so it doesn't end up spewing into our brute force
	 * connection complete messages.
	 */
	hci_reject_con(s, &cr->bdaddr);
	break;

      case NG_HCI_EVENT_COMMAND_COMPL:
	/* Received after most commands - need to handle errors here */
	/* Respect the device telling us to back off the HCI commands */
	if (ccomp->num_cmd_pkts == 0x00) pause = 1;
	
	if (NG_HCI_OGF(ccomp->opcode) == NG_HCI_OGF_LINK_CONTROL) {
	  switch( NG_HCI_OGF(ccomp->opcode) ) {
	    case 0x00:
	      /* Turn commands back on when the device is ready */
	      if (ccomp->num_cmd_pkts > 0x01) {
		pause = 0;
		/* Restart brute force */
		if (brute) {
		  next_bdaddr(&bfaddr);
		  hci_connect(s, &bfaddr);
		}
	      }
	      break;
            case NG_HCI_OCF_PERIODIC_INQUIRY:
	      if (ccomp_result > 0x00) {
		printf("Periodic inquiry failed to start. [Returned: %02x]\n", ccomp_result);
		scanning = 0;
	      }
	      break;
	    case NG_HCI_OCF_EXIT_PERIODIC_INQUIRY:
	      /* No reason we should see this here, since we don't exit this till after this loop */
	      printf("Periodic inquiry unexpectedly died. [Returned: %02x]\n", ccomp_result);
	      scanning = 0;
	  }
	}	 
	break;
  
      case NG_HCI_EVENT_COMMAND_STATUS:
	/* Received after longer commands - need to handle errors here */
	/* Respect the device telling us to back off the HCI commands */
	if (cstat->num_cmd_pkts == 0x00) pause = 1;

	if (NG_HCI_OGF(cstat->opcode) == NG_HCI_OGF_LINK_CONTROL) {
	  switch( NG_HCI_OGF(cstat->opcode) ) {
	    case 0x00:
	      /* Turn commands back on when the device is ready */
	      if (cstat->status == 0x00 && cstat->num_cmd_pkts > 0x01) {
		pause = 0;
		/* Restart brute force */
		if (brute) {
		  next_bdaddr(&bfaddr);
		  hci_connect(s, &bfaddr);
		}
	      }
	      break;
 	    case NG_HCI_OCF_CREATE_CON:
	      if (cstat->status > 0x00) {
		printf("Connection command rejected by device. [Returned: %02x]\n", cstat->status);
		scanning = 0;
	      }	      
 	    case NG_HCI_OCF_DISCON:
	      if (cstat->status > 0x00) {
		printf("Disconnection command rejected by device. [Returned: %02x]\n", cstat->status);
		scanning = 0;
	      }	      
	  }
	}
	break;
    }
  }
  
  hci_periodic_inq_stop(s);

  return 1;
} /* hci_scan */

/* Filters and masks choose what events we see, and what events we want handled for us by the BT stack */
static void set_filter(struct ng_btsocket_hci_raw_filter *filter) {
  memset(filter, 0, sizeof(filter));
  /* This lets us handle the connection requests instead of HC */
  bit_set(filter->packet_mask, NG_HCI_EVENT_CON_REQ - 1);
  /* These choose which events we want to see */
  bit_set(filter->event_mask, NG_HCI_EVENT_COMMAND_COMPL - 1);
  bit_set(filter->event_mask, NG_HCI_EVENT_COMMAND_STATUS - 1);
  bit_set(filter->event_mask, NG_HCI_EVENT_INQUIRY_COMPL - 1);
  bit_set(filter->event_mask, NG_HCI_EVENT_INQUIRY_RESULT - 1);
  bit_set(filter->event_mask, NG_HCI_EVENT_CON_COMPL - 1);
  bit_set(filter->event_mask, NG_HCI_EVENT_CON_REQ - 1);
} /* set_filter */

/* Print the usage message */
static void usage(void)
{
  printf("Usage: ftc -n HCI_node_name [-b]\n"
	 "-n: HCI device, such as ubt0hci\n"
	 "-b: enable brute forcing (must be run as root)\n");
} /* usage */

int main(int argc, char *argv[])
{
  char *node = NULL;
  int s = NULL, n;
  struct ng_btsocket_hci_raw_filter filter;
  int brute = 0;

  /* Set the signal handlers */
  setup_signal_handler(SIGINT, handle_sigint);
  setup_signal_handler(SIGTERM, handle_sigterm);
  setup_signal_handler(SIGQUIT, handle_sigquit);

  /* Process command line arguments */
  while ((n = getopt(argc, argv, "n:b")) != -1) {
    switch (n) {
    case 'n': node = optarg; break;
    case 'b': brute = 1;     break;
    default:
      usage();
      exit(0);
    }
  }

  /* Must have a node name */
  if (node == NULL) {
    usage();
    exit(0);
  }

  printf("Fine Tooth Comb v0.1\n");
  if (brute) printf("Brute force mode enabled.\n");

  /* Create the socket filters */
  set_filter(&filter);

  /* Open a raw HCI socket */
  s = socket_open(node, &filter);

  hci_scan(s, brute);

  /* Clean up */
  close(s);

  return 1;
} /* main */
