/*
 * ftc_util.c
 *
 * Copyright (c) 2003 Bob Fleck <bob@shmoo.com>
 * All rights reserved.
 *
 * Portions of HCI code derived from the FreeBSD hccontrol utility:
 * Copyright (c) 2001-2002 Maksim Yevmenkin <m_evmenkin@yahoo.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "ftc.h"

/* Open a raw HCI socket to Netgraph Bluetooth node */
int socket_open(char const *node, struct ng_btsocket_hci_raw_filter *filter)
{
  struct sockaddr_hci addr;
  int s;
  size_t size;
  
  s = socket(PF_BLUETOOTH, SOCK_RAW, BLUETOOTH_PROTO_HCI);
  if (s < 0) {
    perror("Could not create socket");
    exit(0);
  }
  
  memset(&addr, 0, sizeof(addr));
  addr.hci_len = sizeof(addr);
  addr.hci_family = AF_BLUETOOTH;
  strlcpy(addr.hci_node, node, sizeof(addr.hci_node));

  if (setsockopt(s, SOL_HCI_RAW, SO_HCI_RAW_FILTER, 
		 (void * const)filter, sizeof(*filter)) < 0) {
    perror("Could not setsockopt()");
    exit(0);
  }

  if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    perror("Could not bind socket");
    exit(0);
  }

  if (connect(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    perror("Could not connect socket");
    exit(0);
  }

  return (s);
} /* socket_open */

/* Send HCI data to the unit */
int hci_send(int s, char const *buffer, int size)
{
  int nb, sent = 0;

  assert(buffer != NULL);
  assert(size >= sizeof(ng_hci_cmd_pkt_t));
  assert(size <= sizeof(ng_hci_cmd_pkt_t) + NG_HCI_CMD_PKT_SIZE);
  
  while (sent < size) {
    nb = send(s, (const char *)buffer + sent, size - sent, 0);
    if (nb == -1 && (errno == EAGAIN || errno == EINTR)) continue;
    if (nb <= 0) return nb;
    sent += nb;
  }

  return sent;
} /* hci_send */

/* hci_recv_pkt */
int hci_recv_pkt(int s, char *buffer, int size) {
  int nb, recvd = 0;
  ng_hci_event_pkt_t *e = (ng_hci_event_pkt_t *) buffer;
  
  assert(buffer != NULL);
  assert(size > sizeof(ng_hci_event_pkt_t));
  
 again:
  nb = recv(s, buffer, size, 0);
  if (nb == -1 && (errno == EINTR || errno == EAGAIN)) goto again;
  if (nb < sizeof(ng_hci_event_pkt_t))
    err(EMSGSIZE, "HCI response too small");
  if (nb < 0) return -1;

  if (e->type != NG_HCI_EVENT_PKT) {
    err(EIO, "Request received non-event packet");
  }

  return nb;
} /* hci_recv_pkt */

/* Send HCI request to the unit */
int hci_request(int s, int opcode, char const *cp, int cp_size)
{
  char buffer[512];
  int n, nb;
  ng_hci_cmd_pkt_t *c = (ng_hci_cmd_pkt_t *) buffer;
  ng_hci_event_pkt_t *e = (ng_hci_event_pkt_t *) buffer;
  
  c->type = NG_HCI_CMD_PKT;
  c->opcode = (u_int16_t)opcode;
  c->opcode = htole16(c->opcode);
  
  if (cp != NULL) {
    assert(0 < cp_size && cp_size <= NG_HCI_CMD_PKT_SIZE);
    
    c->length = (u_int8_t)cp_size;
    memcpy(buffer + sizeof(*c), cp, cp_size);
  } else {
    c->length = 0;
  }

  if (hci_send(s, buffer, sizeof(*c) + cp_size) == -1) return -1;
  
  return 1;
} /* hci_request */

/* Send a connection packet to a specified address with standard params */
void hci_connect(int s, bdaddr_t *addr) {
  ng_hci_create_con_cp cp;

  /* Set defaults */
  memset(&cp, 0, sizeof(cp));
  cp.pkt_type = htole16(	NG_HCI_PKT_DM1 | NG_HCI_PKT_DH1 |
				NG_HCI_PKT_DM3 | NG_HCI_PKT_DH3 |
				NG_HCI_PKT_DM5);
  cp.page_scan_rep_mode = NG_HCI_SCAN_REP_MODE0;
  cp.page_scan_mode = NG_HCI_MANDATORY_PAGE_SCAN_MODE;
  cp.clock_offset = 0;
  cp.accept_role_switch = 1;

  memcpy(&cp.bdaddr,addr,sizeof(cp.bdaddr));

  if (hci_request(s, 
		  NG_HCI_OPCODE(NG_HCI_OGF_LINK_CONTROL,NG_HCI_OCF_CREATE_CON),
		  (char const *) &cp, sizeof(cp)) == -1) {
    perror("Failed to send connect packet");
    scanning = 0;
  }
} /* hci_connect */

/* Send Disconnect command to the unit */
void hci_disconnect(int s, int acl)
{
  ng_hci_discon_cp cp;

  /* Build command */
  memset(&cp, 0, sizeof(cp));
  cp.reason = 0x02;
  cp.con_handle = acl;
  
  if (hci_request(s,
		  NG_HCI_OPCODE(NG_HCI_OGF_LINK_CONTROL,NG_HCI_OCF_DISCON),
		  (char const *) &cp, sizeof(cp)) == -1) {
    perror("Failed to send disconnect packet");
    scanning = 0;
  }
} /* hci_disconnect */

/* Reject an incoming connection */
void hci_reject_con(int s, bdaddr_t *addr)
{
  ng_hci_reject_con_cp cp;


  /* Build command */
  memset(&cp, 0, sizeof(cp));
  memcpy(&cp.bdaddr,addr,sizeof(cp.bdaddr));
  cp.reason = FTC_ERR_REJECT;
  
  if (hci_request(s,
		  NG_HCI_OPCODE(NG_HCI_OGF_LINK_CONTROL,NG_HCI_OCF_REJECT_CON),
		  (char const *) &cp, sizeof(cp)) == -1) {
    perror("Failed to send reject_con packet");
    scanning = 0;
  }
} /* hci_reject_con */

/* Start perioidic inquiries */
void hci_periodic_inq_start(int s, int lap) {
  ng_hci_periodic_inquiry_cp cp;

  /* Build command */
  memset(&cp, 0, sizeof(cp));
  /* mult. by 1.28 sec */
  cp.max_period_length = FTC_PI_MAX;
  cp.min_period_length = FTC_PI_MIN;
  cp.inquiry_length = FTC_TIMEOUT_INQR;
  /* Set DIAC LAP for inquiry */
  cp.lap[2] = 0x9e;
  cp.lap[1] = 0x8b;
  /* 0x33 is GIAC, 0x00 is LIAC, 0x00 to 0x3f are valid DIAC */
  cp.lap[0] = lap & 0x3F;
  cp.num_responses = FTC_MAX_INQR;

  /* Send request and expect status response */
  if (hci_request(s,
		  NG_HCI_OPCODE(NG_HCI_OGF_LINK_CONTROL,NG_HCI_OCF_PERIODIC_INQUIRY),
		  (char const *) &cp, sizeof(cp)) == -1) {
    perror("Failed to start periodic inquiry scan");
    scanning = 0;
  }
} /* hci_periodic_inq_start */

/* Abort the periodic inquiry */
void hci_periodic_inq_stop(int s) {
  /* Send request and expect status response */
  if (hci_request(s,
		  NG_HCI_OPCODE(NG_HCI_OGF_LINK_CONTROL,NG_HCI_OCF_EXIT_PERIODIC_INQUIRY),
		  0, 0) == -1) {
    perror("Failed to stop periodic inquire scan - RESET BLUETOOTH DEVICE");
    scanning = 0;
  }
} /* hci_periodic_inq_stop */
